import click
from .lookout import PyLookout


@click.command()
@click.option(
    "--threshold", default=75, help="Percentage threshold for alerting"
)
@click.option(
    "--mode",
    default="local",
    help="Send notifications: simplepush, sendgrid or locally?",
)
@click.option(
    "--containers", is_flag=True, default=False, help="Monitor containers?"
)
@click.option(
    "--background", is_flag=True, default=False, help="Run in background?"
)
def cli(threshold, mode, containers, background):
    """
    Cli interface to easily pass parameters to PyLookout
    """
    lookout = PyLookout(threshold, mode, containers)
    if background:
        lookout.run_in_background()
    else:
        lookout.checker()



import os
import json
from psutil import (
    cpu_count,
    cpu_percent,
    virtual_memory as ram,
    disk_partitions,
    disk_usage,
)
from dataclasses import dataclass, field, InitVar


@dataclass
class Collector:
    # Containers info
    check_containers: InitVar[bool]
    containers: dict = field(default_factory=dict)

    # Hostname
    hostname: str = os.uname()[1]

    # CPU information
    cpu_total: str = cpu_count(logical=True)
    cpu_detail: dict = field(default_factory=dict)
    cpu_percent: str = cpu_percent()

    # RAM information
    ram_total: str = field(default_factory=str)
    ram_avail: str = field(default_factory=str)
    ram_used: str = field(default_factory=str)
    ram_percent: str = ram().percent

    # Disk information
    partitions: list = field(default_factory=list)
    disks_info: dict = field(default_factory=dict)

    def __post_init__(self, check_containers):
        self.cpu_detail = {
            f"Core{i}": p
            for i, p in enumerate(cpu_percent(percpu=True, interval=1))
        }

        self.ram_total = self._convert_bytes(ram().total)
        self.ram_avail = self._convert_bytes(ram().available)
        self.ram_used = self._convert_bytes(ram().used)

        self.partitions = [
            part
            for part in disk_partitions()
            if "loop" not in part.device and "boot" not in part.mountpoint
        ]
        self.disks_info = self._disks_info()
        self.containers = self._get_containers() if check_containers else {}

    def _disks_info(self):
        """
        Get information about each individual partiion,
        except boot and loop partitions.
        """
        dd = {}
        for part in self.partitions:
            du = disk_usage(part.mountpoint)
            disk_info = {
                "du_total": self._convert_bytes(du.total),
                "du_used": self._convert_bytes(du.used),
                "du_free": self._convert_bytes(du.free),
                "du_percent": du.percent,
            }
            dd[part.device] = disk_info
        return dd

    def _inspect_container(self, inspect):
        """
        Read output of 'docker compose <id>' in string format,
        return a dictionary of useful values.
        """
        return {
            "name": inspect["Name"],
            "id": inspect["Id"][0:12],
            "id_full": inspect["Id"],
            "image": inspect["Config"]["Image"],
            "created": inspect["Created"],
            "started": inspect["State"]["StartedAt"],
            "status": inspect["State"]["Status"],
            "ports": inspect["NetworkSettings"]["Ports"],
        }

    def _get_containers(self):
        """
        Get outpus of `docker ps` command,
        and parse all container information from it.
        """
        containers_parsed = {}
        docker_ps = (
            os.popen("docker ps -a | awk '{print $1}'").read().split("\n")
        )
        container_ids = docker_ps[1:-1]
        for container in container_ids:
            inspect = json.loads(
                os.popen(f"docker inspect {container}").read()
            )[0]
            container = self._inspect_container(inspect)
            containers_parsed[container["id"]] = container
        return containers_parsed

    def _convert_bytes(self, bytes, suffix="B"):
        """
        Convert bytes into human readable form
        e.g:
            1253656 => '1.20MB'
            1253656678 => '1.17GB'
        """
        factor = 1024
        for unit in ["", "K", "M", "G", "T", "P"]:
            if bytes < factor:
                return f"{bytes:.2f}{unit}{suffix}"
            bytes /= factor



import logging
from pathlib import Path
from os import getenv
from time import sleep
from urllib import request, parse
from .info_collector import Collector

from sendgrid import SendGridAPIClient
from sendgrid.helpers.mail import Mail, Email, To, Content


class PyLookout:
    def __init__(
        self,
        threshold=75,
        method="sendgrid",
        check_containers=False,
        background=False,
    ):
        logging.basicConfig(
            format="%(asctime)s %(message)s",
            level=logging.INFO,
            handlers=[
                logging.FileHandler(f"{str(Path.home())}/.pylookout.log"),
                logging.StreamHandler(),
            ],
        )

        self.logger = logging.getLogger()
        self.info = Collector(check_containers)
        self.logger.info("Informaton collected successfully!")
        self.critical = threshold
        self.method = method
        self.containers = check_containers
        self.notification = []

    def _messge_percent(self, metric, percent):
        """
        Notification message.
        """
        msg = f"Metric: {metric} ===> Utilization: {percent}%"
        return msg

    def _format_message(self, stage):
        """
        Format notification message.
        """
        total_length = 66
        text_length = 24 + len(self.info.hostname)
        eq = (total_length - text_length) // 2
        return (
            f"{eq*'='}"
            f" pyLookout {stage} on {self.info.hostname} "
            f"{eq*'='}\n"
        )

    def _adjust_message(self):
        """
        Adjust notification message.
        """
        if self.notification != []:
            title = self._format_message("launching")
            ending = self._format_message("finished")
            self.notification.insert(0, title)
            self.notification.append(ending)

    def _simple_push(self):
        """
        Send notifications using Simplepush.
        """
        api_key = getenv("SIMPLEPUSH")
        data = parse.urlencode(
            {
                "key": api_key,
                "title": f"pyLookout on {self.info.hostname}\n",
                "msg": "\n".join(self.notification),
                "event": "event",
            }
        ).encode()
        req = request.Request("https://api.simplepush.io/send", data=data)
        request.urlopen(req)
        self.logger.info("Notification sent successfully!")
        self.logger.info("Notification message:")
        self.logger.info(self.notification)

    def _sendgrid(self):
        """
        Send notifications using SengGrid.
        """
        api_key = getenv("SENDGRID_API_KEY")
        email_from = Email(getenv("SENDGRID_FROM"))
        email_to = To(getenv("SENDGRID_TO"))

        subject = f"pyLookout on {self.info.hostname}\n"
        content = Content("text/plain", "\n".join(self.notification))
        mail = Mail(email_from, email_to, subject, content)

        response = SendGridAPIClient(api_key).client.mail.send.post(
            request_body=mail.get()
        )

        if response.status_code == 202:
            self.logger.info("Email sent succsessfully!")
            self.logger.info("Emailed message:")
            self.logger.info(self.notification)

    def _notify(self):
        """
        Send a notification.
        Available methods:
            * local (print to console)
            * simplepush
            * sendgrid
        """
        self._adjust_message()
        if self.method == "local":
            for notification in self.notification:
                self.logger.info(notification)
        elif self.method == "simplepush":
            self._simple_push()
        elif self.method == "sendgrid":
            self._sendgrid()

    def _containers_status(self, containers):
        """
        Check all container statuses,
        send notifications if monitored container is down.
        """
        for container in containers.values():
            if container["status"] != "running":
                name = container["name"].replace("/", "")
                self.notification.append(
                    f"CONTAINER {name} ({container['id']}) "
                    f"{container['status'].upper()}"
                )
        self.notification[-1] += "\n"

    def _stressed(self, metric, percent):
        """
        Compare a metric with the critical value.
        """
        stressed = True if percent > self.critical else False

        if stressed:
            self.notification.append(self._messge_percent(metric, percent))

    def checker(self):
        """
        One by one check if CPU, RAM and Disk space
        utilization is larger than the critical value.
        """
        self._stressed("CPU", self.info.cpu_percent)
        self._stressed("RAM", self.info.ram_percent)

        for disk in self.info.disks_info.values():
            self._stressed("DISK", disk["du_percent"])

        if self.containers:
            self._containers_status(self.info.containers)

        if self.notification:
            self._notify()

    def run_in_background(self):
        """
        Run checker in background.
        """
        while True:
            self.logger.info("Running checker...")
            self.checker()
            self.logger.info("Checker finished. Sleeping for 60 seconds...")
            sleep(60)
